precision highp float;

#define MAX_PLIGHT 3
#define MAX_SLIGHT 3

#include "material.glsl"						// This include is handled in the shaderloader
#line 7 7

in VertexFragment vertexData;					// Input from vertex stage

out vec4 fragColor;								// Shaded color

uniform vec4 cameraPosition;					// Camera position in world coordinates

uniform bool useDirectionalLight;				// Enables directional light
uniform vec3 directionalLightDirection;			// Directional light's direction
uniform vec3 directionalLightIntensity;			// Directional light's intensity

//PointLight
uniform int pointLightCount;					// Number of point lights used
uniform vec3 pointlightPosition[MAX_PLIGHT];	// Position of point light
uniform vec3 pointlightIntensity[MAX_PLIGHT];	// Intensity of point light
uniform vec3 pointlightAttenuation[MAX_PLIGHT]; // Attenuation of pointlight 1 / (x + dy + d^2z)

//SpotLight
uniform int spotLightCount;						// Number of spot lights used
uniform vec3 spotLightPosition[MAX_SLIGHT];		// Position of spot lights
uniform vec3 spotLightDirection[MAX_SLIGHT];	// Direction of spot light
uniform vec3 spotLightIntensity[MAX_SLIGHT];	// Intensity of spot light
uniform vec3 spotLightAttenuation[MAX_SLIGHT];	// Attenuation of spot light 1 / (x + dy + d^2z)
uniform float spotLightAngle[MAX_SLIGHT];		// Angle of spot light
uniform float spotLightAngularAttenuation[MAX_SLIGHT];	// Angular attenuation of spot light

vec4 material(VertexFragment fragment, vec3 diffuseIntensity, vec3 specularIntensity);		// Prototype for user function
FragmentIllumination fragment(VertexFragment fragment);		// Prototype for user function
float linInt(float edge0, float edge1, float x);


void main()
{
	VertexFragment vertex = vertexData;
	vertex.position /= vertex.position.w;
	vertex.ndc /= vertex.ndc.w;
	vertex.normal = normalize(vertex.normal);
	vertex.tangent = normalize(vertex.tangent);

	FragmentIllumination fragm = fragment(vertex);
	vertex.normal = fragm.normal;

	vec3 view = normalize(vertex.position.xyz - cameraPosition.xyz);

	vec3 diffuseIntensity, specularIntensity;

	if (isShadeless)
	{
		diffuseIntensity = vec3(1.0, 1.0, 1.0);
		specularIntensity = vec3(0.0, 0.0, 0.0);
	}
	else
	{
		diffuseIntensity = vec3(0.0, 0.0, 0.0);
		specularIntensity = vec3(0.0, 0.0, 0.0);

		//Directional light
		if (useDirectionalLight)
		{
			vec3 r = reflect(-directionalLightDirection, vertex.normal);
			diffuseIntensity += abs(dot(fragm.normal, -directionalLightDirection)) * directionalLightIntensity;
			specularIntensity += pow(max(0.0, dot(view, r)), specularHardness) * directionalLightIntensity;
		}

		//Point lights
		for (int i = 0; i < pointLightCount; i++)
		{
			vec3 pointlightDirection = pointlightPosition[i].xyz - vertex.position.xyz;
			float d = length(pointlightDirection);
			pointlightDirection = normalize(pointlightDirection);

			float att = 1.0 / (pointlightAttenuation[i].x + pointlightAttenuation[i].y * d + pointlightAttenuation[i].z * d * d);
			vec3 r = reflect(pointlightDirection, fragm.normal);

			diffuseIntensity += abs(dot(fragm.normal, pointlightDirection)) * pointlightIntensity[i] * att;
			specularIntensity += pow(max(0.0, dot(view, r)), fragm.hardness) * pointlightIntensity[i] * att;
		}

		//Spot
		for (int i = 0; i < spotLightCount; i++)
		{
			vec3 l = vertex.position.xyz - spotLightPosition[i];
			vec3 lnormal = normalize(l);
			float angle = dot(lnormal, spotLightDirection[i]);

			if (angle >= cos(spotLightAngle[i])) // Otherwise we are outside the cone
			{
				float d = length(l);
				float att = 1.0 / (spotLightAttenuation[i].x + spotLightAttenuation[i].y * d + spotLightAttenuation[i].z * d * d);
				float angAtt = 1.0 - linInt((1.0 - spotLightAngularAttenuation[i]) * spotLightAngle[i], spotLightAngle[i], acos(angle));

				vec3 r = reflect(-spotLightDirection[i], fragm.normal);

				diffuseIntensity += abs(dot(fragm.normal, -spotLightDirection[i])) * spotLightIntensity[i] * att * angAtt;
				specularIntensity += pow(max(0.0, dot(view, r)), fragm.hardness) * spotLightIntensity[i] * att * angAtt;
			}
		}
	}

	fragColor = material(vertex, diffuseIntensity, specularIntensity);
}


//----------------------------------
//Helper methods
//----------------------------------
float linInt(float edge0, float edge1, float x)
{
	return clamp((x - edge0) / (edge1 - edge0), 0.0, 1.0);
}