precision highp float;

#include "../MaterialShader/material.glsl"			// This includes are handled in the shaderloader
#include "octree_common.glsl"
#line 6 6

uniform layout(binding = 10) sampler2D deferredNormal;
uniform layout(binding = 11) sampler2D deferredTangent;
uniform layout(binding = 12) sampler2D deferredDepth;

uniform layout(binding = 0) sampler3D brickTexture;
readonly uniform layout(size1x32, binding = 2) uimageBuffer lockBuffer;

layout (std430, binding = 1) buffer linkBuffer
{
	LinkBufferEntry entry[];
};

#include "octree_traverse.glsl" //Has to be included after buffer definition

#line 22 22

uniform uint steps;
uniform vec3 cameraPosition;

in vec2 texCoord;

layout(location = 0) out vec3 diffuseIllumination;
layout(location = 1) out vec3 specularIllumination;


vec4 CalculateWorldPos(float _depth)
{
	vec4 position = vec4(texCoord.x * 2.0 - 1.0, texCoord.y * 2.0 - 1.0, _depth * 2.0 - 1.0, 1.0);
	position = inverseCameraMatrix * position;
	position = position / position.w;
	return position;
}



const uint coneStart = 0;
const uint coneEnd = 6;
vec3 coneDirections[7] =
{
	vec3(0,0,1),
	vec3(0,     0.866,  0.75),
	vec3(0.8236,  0.2676,  0.75),
	vec3(0.5090,  -0.7006, 0.75),
	vec3(0.5090,     -0.7006, 0.75),
	vec3(-0.8236, 0.2676, 0.75),
	vec3(0,0,0)
};


mat3 tbn;

vec4 traceCone(vec3 _direction, float _angle, vec3 _start, uint _steps, float _stepSize, float _stepAdaption)
{
	vec3 rayDirection = normalize(_direction);

	vec3 rayPosition = _start;

	rayDirection *=  (_stepSize * (2.0/256.0));
	float rayStepLength = length(_start - rayPosition);

	float alpha = 0.0;
	vec3 color = vec3(0,0,0);

	for (uint i = 1; i <= _steps; i++)
	{
		vec3 abspos = abs(rayPosition);

		if (abspos.x <= 1.0 && abspos.y <= 1.0 && abspos.z <= 1.0) //Point is in octree
		{
			//Calculate diameter and level
			float diam = length(rayPosition - _start) * _angle; // == / (2/512) * 2
			float optimal_level = log2(diam);

			//int level_l = int(clamp(9 - ceil(optimal_level), 0, 8));
			int level_h = int(clamp(9 - floor(optimal_level), 0, 8));
			int level_l = level_h - 1;
			float level_interpol = clamp((diam - pow(2, 9-level_h)) / (pow(2, 9-level_l) - pow(2, 9-level_h)), 0, 1);

			uint hidx = 0;
			uint lidx = 0;
			vec3 halfSize;
			vec3 hcenter;
			vec3 lcenter;

			traverse(rayPosition, level_l, lcenter, halfSize, lidx);
			hcenter = lcenter;
			bool foundLevel = traverseLevel(rayPosition, lidx, hcenter, halfSize, hidx);

				
			if (foundLevel)
			{
				vec4 opacity_h = vec4(0,0,0,0);
				vec4 opacity_l = vec4(0,0,0,0);

				//Load values
				if (hidx != 0)
				{
					vec3 brick_pos = posInBrick(rayPosition, hidx, hcenter, halfSize);

					opacity_h = texture(brickTexture, brick_pos);
					//opacity_h.a = 1.0 - pow((1.0 - opacity_h.a), (rayStepLength * 7.0 * float(pow(2.0, level_h-2))));
					opacity_h.rgb *= opacity_h.a;
					//opacity_h.a = 1.0 - pow((1.0 - opacity_h.a), (rayStepLength * float(pow(2.0, level_h-2))));
					//opacity_h.a = 1.0 - pow((1.0 - opacity_h.a), (rayStepLength * 4.0 * float(pow(2.0, level_h-2))));
				}
				if (lidx != 0)
				{
					vec3 brick_pos = posInBrick(rayPosition, lidx, lcenter, halfSize * 2.0);

					opacity_l = texture(brickTexture, brick_pos);
					//opacity_l.a = 1.0 - pow((1.0 - opacity_l.a), (rayStepLength * 7.0 * float(pow(2.0, level_l-2))));
					opacity_l.rgb *= opacity_l.a;
					//opacity_l.a = 1.0 - pow((1.0 - opacity_l.a), (rayStepLength * 4.0 * float(pow(2.0, level_l-2))));
				}

				//Interpolation between levels				
				vec4 opacity = level_interpol * opacity_l + (1.0 - level_interpol) * opacity_h;
				//opacity.a = 1.0 - pow((1.0 - opacity.a), (rayStepLength * float(pow(2.0, level_h-1))));

				//opacity.a *= (1.0 / (1.0 + 1.0 * length(_start - rayPosition)));

				color = alpha * color + (1.0 - alpha) * opacity.rgb;
				alpha = alpha + (1 - max(alpha, 0)) * opacity.a;
			}
		}

		rayPosition += rayDirection;
		rayDirection *= _stepAdaption;
		rayStepLength = length(rayDirection);
	}

	return vec4(color, alpha);
}

void main()
{
	//float angle = tan(0.4363) * 512.0;
	float angle = tan(0.5235) * 512.0;
	//float angle = tan(0.3) * 512.0;

	//Read geometry information from deffered textures
	vec4 fragNormal = texture(deferredNormal, texCoord.xy);
	float fragDepth = texture(deferredDepth, texCoord.xy).x;

	if (fragNormal.xyz == vec3(0,0,0))
	{
		discard;
		return;
	}

	vec4 fragPosition = CalculateWorldPos(fragDepth);
	vec3 tangent = texture(deferredTangent, texCoord).xyz;
	vec3 bitangent = normalize(cross(tangent, fragNormal.xyz));
	gl_FragDepth = fragDepth;

	tbn = mat3(tangent, bitangent, fragNormal.xyz);
	float alphaTotal = 0.0;
	vec3 diffuseTotal = vec3(0,0,0);

	vec3 rayStart = (cameraMatrixArray[2] * fragPosition).xyz;

	//Diffuse Cones
	float coneWeight = 1.0 / float(coneEnd - coneStart);
	for (uint coneId = coneStart; coneId < coneEnd; coneId++)
	{
		vec4 result = traceCone((tbn * coneDirections[coneId]) * 1.0 , angle, rayStart + fragNormal.xyz * (2.0/256.0) * 1.0, steps, 0.2, 1.0);

		alphaTotal += coneWeight * 1.0 * max(0, result.a);
		diffuseTotal += coneWeight * 1.0 * result.rgb;
	}

	//Specular Cone
	float specAngle = max(0.1, acos(pow(0.1, 1.0 / fragNormal.a)));
	vec3 view = normalize(fragPosition.xyz - cameraPosition);
	vec3 specDir = normalize(reflect(view, fragNormal.xyz));
	vec4 specular = traceCone(specDir, tan(specAngle) * 512.0, rayStart + fragNormal.xyz * (2.0/256.0) * 1.0, steps * 2, 0.15, 1.0);

	diffuseIllumination = diffuseTotal.rgb + 0.5 * specular.rgb;
	//specularIllumination = 0.5 * specular.rgb;	

	//diffuseIllumination = vec3(0,0,0);
	specularIllumination = vec3(0,0,0);
}