precision highp float;

#include "../MaterialShader/material.glsl"			// This includes are handled in the shaderloader
#include "octree_common.glsl"
#line 6 6

uniform layout(binding = 3) sampler2D deferredNormal;
uniform layout(binding = 4) sampler2D deferredTangent;
uniform layout(binding = 5) sampler2D deferredDepth;

uniform layout(binding = 0) sampler3D brickTexture;
readonly uniform layout(size1x32, binding = 2) uimageBuffer lockBuffer;

layout (std430, binding = 1) buffer linkBuffer
{
	LinkBufferEntry entry[];
};

#include "octree_traverse.glsl" //Has to be included after buffer definition

uniform uint steps;

in vec2 texCoord;

out vec4 fragColor;					// Shaded color


vec4 CalculateWorldPos(float _depth)
{
	vec4 position = vec4(texCoord.x * 2.0 - 1.0, texCoord.y * 2.0 - 1.0, _depth * 2.0 - 1.0, 1.0);
	position = inverseCameraMatrix * position;
	position = position / position.w;
	return position;
}



const uint coneStart = 0;
const uint coneEnd = 6;
vec3 coneDirections[7] =
{
	vec3(0,0,1),
	vec3(0,     0.866,  0.5),
	vec3(0.8236,  0.2676,  0.5),
	vec3(0.5090,  -0.7006, 0.5),
	vec3(0.5090,     -0.7006, 0.5),
	vec3(-0.8236, 0.2676, 0.5),
	vec3(0.0, 0.0, 0.0)
};

void main()
{
	const float stepWidth = 0.5;
	const float startDist = 0.25;
	float angle = tan(0.5235) * 512.0;

	//Read geometry information from deffered textures
	vec3 fragNormal = texture(deferredNormal, texCoord.xy).xyz;
	float fragDepth = texture(deferredDepth, texCoord.xy).x;

	if (fragNormal.xyz == vec3(0,0,0))
	{
		discard;
		return;
	}

	vec4 fragPosition = CalculateWorldPos(fragDepth);
	vec3 tangent = texture(deferredTangent, texCoord).xyz;
	vec3 bitangent = normalize(cross(tangent, fragNormal));
	gl_FragDepth = fragDepth;

	mat3 tbn = mat3(tangent, bitangent, fragNormal);
	float alphaTotal = 0.0;

	vec3 rayStart = (cameraMatrixArray[2] * fragPosition).xyz + fragNormal.xyz * (2.0/256.0) * startDist;
	float coneWeight = 2.0 / float(coneEnd - coneStart);

	for (uint coneId = coneStart; coneId < coneEnd; coneId++)
	{
		vec3 rayDirection = normalize(tbn * coneDirections[coneId]);

		vec3 rayPosition = rayStart;

		rayDirection *=  (stepWidth * (2.0/256.0));
		float rayStepLength = length(rayStart - rayPosition);

		float alpha = 0.0;

		for (uint i = 1; i <= steps; i++)
		{
			vec3 abspos = abs(rayPosition);

			if (abspos.x <= 1.0 && abspos.y <= 1.0 && abspos.z <= 1.0) //Point is in octree
			{
				//Calculate diameter and level
				float diam = length(rayPosition - rayStart) * angle; // == / (2/512) * 2
				float optimal_level = log2(diam);

				int level_l = int(clamp(9 - ceil(optimal_level), 0, 8));
				int level_h = int(clamp(9 - floor(optimal_level), 0, 8));
				float level_interpol = clamp((diam - pow(2, 9-level_h)) / (pow(2, 9-level_l) - pow(2, 9-level_h)), 0, 1);

				uint hidx = 0;
				uint lidx = 0;
				vec3 halfSize;
				vec3 hcenter;
				vec3 lcenter;

				traverse(rayPosition, level_l, lcenter, halfSize, lidx);
				hcenter = lcenter;
				bool foundLevel = traverseLevel(rayPosition, lidx, hcenter, halfSize, hidx);

				
				if (foundLevel)
				{
					float opacity_h = 0.3;
					float opacity_l = 0.3;

					//Load values
					if (hidx != 0)
					{
						vec3 brick_pos = posInBrick(rayPosition, hidx, hcenter, halfSize);

						opacity_h = max(0, texture(brickTexture, brick_pos).a);
						opacity_h = 1.0 - pow((1.0 - opacity_h), (rayStepLength * float(pow(2.0, level_h-1))));
					}
					if (lidx != 0)
					{
						vec3 brick_pos = posInBrick(rayPosition, lidx, lcenter, halfSize * 2.0);

						opacity_l = max(0, texture(brickTexture, brick_pos).a);
						opacity_l = 1.0 - pow((1.0 - opacity_l), (rayStepLength * float(pow(2.0, level_l-1))));
					}

					//Interpolation between levels				
					float opacity = level_interpol * opacity_l + (1.0 - level_interpol) * opacity_h;

					opacity *= (1.0 / (1.0 + 0.0 * length(rayStart - rayPosition)));
					alpha = alpha + (1 - max(alpha, 0)) * opacity;
				}
			}

			rayPosition += rayDirection;
			rayStepLength = length(rayDirection);
		}

		alphaTotal += coneWeight * 1.1 * max(0, alpha);
	}

	float val = 1.0 - (alphaTotal * 1.0);
	fragColor = vec4(val, val, val, 1);
}