#include "octree_common.glsl"

#line 3 3

coherent uniform layout(size1x32, binding = 2) uimageBuffer lockBuffer;

layout (std430, binding = 1) buffer linkBuffer
{
	LinkBufferEntry entry[];
};

layout(binding = 0, offset = 0) uniform atomic_uint idxCounter;
layout(binding = 2, offset = 0) uniform atomic_uint levelCounter;

layout (std430, binding = 3) buffer deferredBuffer
{
	DeferredBufferEntry defEntry[];
};

uniform int maxLevel;
uniform int leafLevel;


uint addToUnorm(uint _value, uint _toAdd)
{
	vec3 val = unpackUnorm4x8(_value).xyz;
	uint count = bitfieldExtract(_value, 24, 8);
	val = (count * val + unpackUnorm4x8(_toAdd).xyz) / (count + 1);
	count++;
	uint result = packUnorm4x8(vec4(val, 0));
	result = bitfieldInsert(result, count, 24, 8);
	return result;
}

uint addToSnorm(uint _value, uint _toAdd)
{
	vec3 val = unpackSnorm4x8(_value).xyz;
	uint count = bitfieldExtract(_value, 24, 8);
	val = (count * val + unpackSnorm4x8(_toAdd).xyz) / (count + 1);
	count++;
	uint result = packSnorm4x8(vec4(val, 0));
	result = bitfieldInsert(result, count, 24, 8);
	return result;
}


void main()
{	
	DeferredBufferEntry data = defEntry[gl_VertexID];
	vec3 pos = unpackSnorm3x10(data.pos);

	pos += vec3(-0.0001, -0.0001, -0.0001) * sign(unpackSnorm4x8(data.normal).xyz); //This stabilizes the splitting decision, reason: 10 bits can only store 1024 values
											//-> values fall often exactly on the center of a voxel

	vec3 halfSize = vec3(1,1,1);
	vec3 center = vec3(0,0,0);

	uint idx = 0;
	uint parentIdx = 0;
	int octidx = 0;
	int level;

	//Try to reach current level
	for (level = 0; level < maxLevel-1; level++)
	{		
		//Calculate position in current octree level
		vec3 octdir = pos - center;
		ivec3 octvec = max(ivec3(0,0,0), ivec3(sign(octdir)));
		octidx = octvec.x * 4 + octvec.y * 2 + octvec.z;
		parentIdx = idx;
		idx = entry[idx].children + octidx;

		//Calculate extends for next level
		halfSize /= 2.0;
		center += halfSize * sign(octdir);
	}

	if (maxLevel < leafLevel) //If not leafLevel -> Check if allocate is needed
	{
		uint lockVal = imageAtomicCompSwap(lockBuffer, int(idx), 0U, 1U);

		if (lockVal == 0)
		{
			uint nextIdx = 1 + atomicCounterIncrement(idxCounter) * 8;
			atomicCounterIncrement(levelCounter);

			entry[idx].children = nextIdx;
			imageAtomicExchange(lockBuffer, int(idx), 2U);

			//Clear buffers & set internal ptr
			for (int i = 0; i < 8; i++)
			{
				imageStore(lockBuffer, int(nextIdx) + i, uvec4(0U,0U,0U,0U));
			}

			if (maxLevel < leafLevel - 1)
			{
				entry[nextIdx + 0] = LinkBufferEntry(0U,nextIdx + 4,0U,nextIdx + 2,0U,nextIdx + 1,0U);
				entry[nextIdx + 1] = LinkBufferEntry(0U,nextIdx + 5,0U,nextIdx + 3,0U,0U,nextIdx + 0);
				entry[nextIdx + 2] = LinkBufferEntry(0U,nextIdx + 6,0U,0U,nextIdx + 0,nextIdx + 3,0U);
				entry[nextIdx + 3] = LinkBufferEntry(0U,nextIdx + 7,0U,0U,nextIdx + 1,0U,nextIdx + 2);
				entry[nextIdx + 4] = LinkBufferEntry(0U,0U,nextIdx + 0,nextIdx + 6,0U,nextIdx + 5,0U);
				entry[nextIdx + 5] = LinkBufferEntry(0U,0U,nextIdx + 1,nextIdx + 7,0U,0U,nextIdx + 4);
				entry[nextIdx + 6] = LinkBufferEntry(0U,0U,nextIdx + 2,0U,nextIdx + 4,nextIdx + 7,0U);
				entry[nextIdx + 7] = LinkBufferEntry(0U,0U,nextIdx + 3,0U,nextIdx + 5,0U,nextIdx + 6);
			}
			else
			{
				for (int i = 0; i < 8; i++)
					entry[nextIdx + i] = LinkBufferEntry(0U, 0U, 0U, 0U, 0U, 0U, 0U);
			}
		}
	}
	else if (maxLevel == leafLevel && level >= leafLevel - 1) //Otherwise store data
	{
		//Convert position to 10 bit uint
		entry[idx].children = packSnorm3x10(vec3(center));

		uint insert, last, current;
		insert = 0;
		last = 1;

		while (last != current)
		{
			last = entry[idx].xplus;
			insert = addToSnorm(last, data.normal);
			current = atomicCompSwap(entry[idx].xplus, last, insert);
		}

		insert = 0;
		last = 1;

		while (last != current)
		{
			last = entry[idx].xminus;
			insert = addToUnorm(last, data.diffuse);
			current = atomicCompSwap(entry[idx].xminus, last, insert);
		}

		insert = 0;
		last = 1;

		while (last != current)
		{
			last = entry[idx].yplus;
			insert = addToUnorm(last, data.opacityHardness);
			current = atomicCompSwap(entry[idx].yplus, last, insert);
		}
			
		vec3 emit = unpackUnorm4x8(data.emitting).rgb;

		entry[idx].yminus = uint(emit.x * 65535.0);
		entry[idx].zplus = uint(emit.y * 65535.0);
		entry[idx].zminus = uint(emit.z * 65535.0);
		
		imageStore(lockBuffer, int(idx), uvec4(2U, 0U, 0U, 0U));
	}
}